import arviz as az
import pymc3 as pm
import numpy as np
import ujson as json

with open("radon.json", "rb") as f:
    radon_data = json.load(f)

radon_data = {
    key: np.array(value) if isinstance(value, list) else value
    for key, value in radon_data.items()
}

coords = {
    "Level": ["Basement", "Floor"],
    "obs_id": np.arange(radon_data["y"].size),
    "County": radon_data["county_name"],
    "g_coef": ["intercept", "slope"],
}

with pm.Model(coords=coords):
    floor_idx = pm.Data("floor_idx", radon_data["x"], dims="obs_id")
    county_idx = pm.Data("county_idx", radon_data["county"], dims="obs_id")
    uranium = pm.Data("uranium", radon_data["u"], dims="County")

    # Hyperpriors:
    g = pm.Normal("g", mu=0.0, sigma=10.0, dims="g_coef")
    sigma_a = pm.Exponential("sigma_a", 1.0)

    # Varying intercepts uranium model:
    a = pm.Deterministic("a", g[0] + g[1] * uranium, dims="County")
    za_county = pm.Normal("za_county", mu=0.0, sigma=1.0, dims="County")
    a_county = pm.Deterministic("a_county", a + za_county * sigma_a, dims="County")
    # Common slope:
    b = pm.Normal("b", mu=0.0, sigma=1.0)

    # Expected value per county:
    theta = a_county[county_idx] + b * floor_idx
    # Model error:
    sigma = pm.Exponential("sigma", 1.0)

    y = pm.Normal("y", theta, sigma=sigma, observed=radon_data["y"], dims="obs_id")

    prior = pm.sample_prior_predictive(500)
    trace = pm.sample(
        1000, tune=500, target_accept=0.99, random_seed=117
    )
    post_pred = pm.sample_posterior_predictive(trace)
    idata = az.from_pymc3(trace, prior=prior, posterior_predictive=post_pred)



# pylint: disable=wildcard-import,wrong-import-position
"""Base ArviZ features and converters."""

import logging

_log = logging.getLogger(__name__)

from .base import generate_dims_coords, dict_to_dataset, make_attrs, ndarray_to_dataarray
from .converters import *
from .datasets import load_arviz_data, list_datasets, get_data_home, clear_data_home
from .io_cmdstanpy import from_cmdstanpy
from .io_dict import from_dict
from .io_emcee import from_emcee
from .rcparams import rcParams, rc_context
from .sel_utils import *
from ._version import __version__



"""ArviZ testing utilities."""

from typing import Dict, List, Tuple, Union

from datatree import DataTree


def check_multiple_attrs(
    test_dict: Dict[str, List[str]], parent: DataTree
) -> List[Union[str, Tuple[str, str]]]:
    """Perform multiple hasattr checks on InferenceData objects.

    It is thought to first check if the parent object contains a given dataset,
    and then (if present) check the attributes of the dataset.

    Given the output of the function, all mismatches between expectation and reality can
    be retrieved: a single string indicates a group mismatch and a tuple of strings
    ``(group, var)`` indicates a mismatch in the variable ``var`` of ``group``.

    Parameters
    ----------
    test_dict: dict of {str : list of str}
        Its structure should be `{dataset1_name: [var1, var2], dataset2_name: [var]}`.
        A ``~`` at the beginning of a dataset or variable name indicates the name NOT
        being present must be asserted.
    parent: InferenceData
        InferenceData object on which to check the attributes.

    Returns
    -------
    list
        List containing the failed checks. It will contain either the dataset_name or a
        tuple (dataset_name, var) for all non present attributes.

    Examples
    --------
    The output below indicates that ``posterior`` group was expected but not found, and
    variables ``a`` and ``b``:

        ["posterior", ("prior", "a"), ("prior", "b")]

    Another example could be the following:

        [("posterior", "a"), "~observed_data", ("sample_stats", "~log_likelihood")]

    In this case, the output indicates that variable ``a`` was not found in ``posterior``
    as it was expected, however, in the other two cases, the preceding ``~`` (kept from the
    input negation notation) indicates that ``observed_data`` group should not be present
    but was found in the InferenceData and that ``log_likelihood`` variable was found
    in ``sample_stats``, also against what was expected.

    """
    failed_attrs: List[Union[str, Tuple[str, str]]] = []
    for dataset_name, attributes in test_dict.items():
        if dataset_name.startswith("~"):
            if hasattr(parent, dataset_name[1:]):
                failed_attrs.append(dataset_name)
        elif hasattr(parent, dataset_name):
            dataset = getattr(parent, dataset_name)
            for attribute in attributes:
                if attribute.startswith("~"):
                    if hasattr(dataset, attribute[1:]):
                        failed_attrs.append((dataset_name, attribute))
                elif not hasattr(dataset, attribute):
                    failed_attrs.append((dataset_name, attribute))
        else:
            failed_attrs.append(dataset_name)
    return failed_attrs



"""Dictionary specific conversion code."""
import warnings
from typing import Optional

from datatree import DataTree

from .base import dict_to_dataset
from .rcparams import rcParams


def from_dict(
    data,
    *,
    name=None,
    sample_dims=None,
    save_warmup=None,
    index_origin: Optional[int] = None,
    coords=None,
    dims=None,
    pred_dims=None,
    pred_coords=None,
    check_conventions=True,
    attrs=None,
):
    """Convert nested dictionary into a DataTree.

    It uses :func:`dict_to_dataset` to convert a nested dictionary to a DataTree
    using ArviZ conventions.

    Parameters
    ----------
    data : dict of {str : dict of {hashable: array_like}}
        Dictionary to convert to DataTree. It must be a nested dictionary.
        The keys of the outer dictionary are group names, the values
        of the outer dictionary must be dictionaries themselves with the
        variables that should be stored in that group. These inner dictionaries
        are passed to :func:`dict_to_dataset`.

        Depending on the group name, the arguments used when calling `dict_to_dataset`
        can take different arguments from their defaults:

        * If ``data`` is a substring of the group name, ``sample_dims`` will be
          set to an empty list: ``[]``
        * If ``predictions`` is a substring of the group name, `pred_coords`
          and `pred_dims` will be passed to :func:`dict_to_dataset` as ``coords``
          and ``dims``.
        * If ``log_likelihood`` is a substring of the group name,
          ``skip_event_dims`` is set to ``True``.

    name : str, optional
        Name of the DataTree root node. This is used as the DataTree name by
        ArviZ.
    sample_dims : iterable of hashable, optional
        Dimensions that should be assumed to be present in _all_ variables.
        If missing, they will be added as the dimensions corresponding to the
        leading axis.
    save_warmup : bool, optional
        Save warmup iterations DataTree. If not defined, use default
        defined by the rcParams. When set to ``False``, groups in `data` that have
        ``warmup`` as substring will be ignored.
    index_origin : int, optional
        Start value to use default integer ids for dimensions without provided
        coordinate values. Defaults to ``data.index_origin``.
    coords : dict of {hashable : list}, optional
        A dictionary containing the values that are used as index. The key
        is the name of the dimension, the values are the index values.
    dims : dict of {hashable : list of hashable}, optional
        A mapping from variable names to a list of dimension names for the variable.
    pred_dims : dict of {str : list of str}, optional
        A mapping from variables to a list of coordinate names for predictions.
    pred_coords : dict of {str : list}, optional
        A mapping from variables to a list of coordinate values for predictions.
    check_conventions : bool, default True
        Check some ArviZ conventions on dimension meaning.
    attrs : dict of {str : dict}, optional
        A dictionary containing attributes for different groups. Its keys should
        match keys in `data`, with the exception of "/" which is used to set global
        attributes.

    Returns
    -------
    DataTree

    See Also
    --------
    datatree.DataTree.from_dict
    """
    d_ds = {}
    if attrs is None:
        attrs = {}

    bad_attr_groups = [group for group in attrs if (group not in data) and (group != "/")]
    if bad_attr_groups:
        warnings.warn(
            f"Found groups in `attrs` ({bad_attr_groups}) that are not present in `d`. "
            "They will be ignored",
            UserWarning,
        )

    if save_warmup is None:
        save_warmup = rcParams["data.save_warmup"]
    for group, group_data in data.items():
        if save_warmup is False and "warmup" in group:
            continue
        d_ds[group] = dict_to_dataset(
            group_data,
            attrs=attrs.get(group, None),
            coords=pred_coords if "predictions" in group else coords,
            dims=pred_dims if "predictions" in group else dims,
            sample_dims=[] if "data" in group else sample_dims,
            index_origin=index_origin,
            skip_event_dims="log_likelihood" in group,
            check_conventions=check_conventions,
        )
    dt = DataTree.from_dict(d_ds, name=name)
    if "/" in attrs:
        dt.attrs = attrs["/"]
    return dt



"""ArviZ type definitions."""
from collections.abc import Hashable, Mapping
from typing import TYPE_CHECKING, Any, Iterable

if TYPE_CHECKING:
    from numpy.typing import ArrayLike
    from xarray.core.types import DaCompatible

CoordSpec = Mapping[Hashable, Any]
DimSpec = Mapping[Hashable, Iterable[Hashable]]

DictData = Mapping[Hashable, "ArrayLike"]
